﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Threading;
using HEAL.Attic;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Optimization.Operators;
using HeuristicLab.Parameters;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic.Regression {
  [StorableType("4318C6BD-E0A1-45FE-AC30-96E7F73B51FB")]
  public class SymbolicRegressionConstraintAnalyzer : SymbolicDataAnalysisAnalyzer, ISymbolicExpressionTreeAnalyzer {
    private const string ProblemDataParameterName = "ProblemData";
    private const string ConstraintViolationParameterName = "ConstraintViolations";
    private const string ConstraintUnsatisfiedSolutionsParameterName = "Constraint Unsatisfied Solutions";

    #region parameter properties

    public ILookupParameter<IRegressionProblemData> RegressionProblemDataParameter =>
      (ILookupParameter<IRegressionProblemData>) Parameters[ProblemDataParameterName];

    public IResultParameter<DataTable> ConstraintViolationParameter =>
      (IResultParameter<DataTable>) Parameters[ConstraintViolationParameterName];

    public IResultParameter<DataTable> ConstraintUnsatisfiedSolutionsParameter =>
      (IResultParameter<DataTable>)Parameters[ConstraintUnsatisfiedSolutionsParameterName];

    #endregion

    public override bool EnabledByDefault => false;
    public static int Iterations { get; set; } = 0;

    public IBoundsEstimator BoundsEstimator { get; set; } = new IABoundsEstimator();

    [StorableConstructor]
    protected SymbolicRegressionConstraintAnalyzer(StorableConstructorFlag _) : base(_) { }

    protected SymbolicRegressionConstraintAnalyzer(SymbolicRegressionConstraintAnalyzer original, Cloner cloner) :
      base(original, cloner) { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new SymbolicRegressionConstraintAnalyzer(this, cloner);
    }

    public SymbolicRegressionConstraintAnalyzer() {
      Parameters.Add(new LookupParameter<IRegressionProblemData>(ProblemDataParameterName,
        "The problem data of the symbolic data analysis problem."));
      Parameters.Add(new ResultParameter<DataTable>(ConstraintViolationParameterName,
        "Shows the number of constraint violations!"));
      Parameters.Add(new ResultParameter<DataTable>(ConstraintUnsatisfiedSolutionsParameterName,
        "Shows the number of solutions with unsatisfied constraints."));


      ConstraintViolationParameter.DefaultValue = new DataTable(ConstraintViolationParameterName) {
        VisualProperties = {
          XAxisTitle = "Generations",
          YAxisTitle = "Constraint Violations"
        }
      };

      ConstraintUnsatisfiedSolutionsParameter.DefaultValue = new DataTable(ConstraintUnsatisfiedSolutionsParameterName) {
        VisualProperties = {
          XAxisTitle = "Generations",
          YAxisTitle = "Constraint Unsatisfied Solutions"
        }
      };
    }

    public override void InitializeState() {
      Iterations = 0;
      base.InitializeState();
    }

    public override void ClearState() {
      Iterations = 0;
      base.ClearState();
    }


    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      if (Parameters.ContainsKey(ConstraintUnsatisfiedSolutionsParameterName)) return;
      Parameters.Add(new ResultParameter<DataTable>(ConstraintUnsatisfiedSolutionsParameterName,
        "Shows the number of solutions with unsatisfied constraints."));

      ConstraintUnsatisfiedSolutionsParameter.DefaultValue = new DataTable(ConstraintUnsatisfiedSolutionsParameterName) {
        VisualProperties = {
          XAxisTitle = "Generations",
          YAxisTitle = "Constraint Unsatisfied Solutions"
        }
      };
    }

    public override IOperation Apply() {
      Iterations++;

      var problemData = RegressionProblemDataParameter.ActualValue;
      var trees = SymbolicExpressionTreeParameter.ActualValue;

      var results = ResultCollectionParameter.ActualValue;
      var constraints = problemData.IntervalConstraints.EnabledConstraints;
      var variableRanges = problemData.VariableRanges.GetReadonlyDictionary();
      var intervalCollection = problemData.VariableRanges;
      var newDataTable = ConstraintViolationParameter.ActualValue;
      var solutions = this.SymbolicExpressionTree.ToArray();

      if (newDataTable.Rows.Count == 0)
        foreach (var constraint in constraints)
          newDataTable.Rows.Add(new DataRow(constraint.Expression));

      foreach (var constraint in constraints) {
        var violations = trees.Count(tree => IntervalUtil.IntervalConstraintViolation(constraint, BoundsEstimator, intervalCollection, tree) > 0.0);

        newDataTable.Rows[constraint.Expression].Values.Add(violations);
      }

      var constraintUnsatisfiedSolutionsDataTable = ConstraintUnsatisfiedSolutionsParameter.ActualValue;
      if (constraintUnsatisfiedSolutionsDataTable.Rows.Count == 0)
        constraintUnsatisfiedSolutionsDataTable.Rows.Add(new DataRow(ConstraintUnsatisfiedSolutionsParameterName));

      constraintUnsatisfiedSolutionsDataTable.Rows[ConstraintUnsatisfiedSolutionsParameterName]
        .Values
        .Add(solutions.Count(s => IntervalUtil.IntervalConstraintsViolation(constraints, BoundsEstimator, intervalCollection, s).Any(x => x != 0.0)));

      return base.Apply();
    }

    
  }
}